@@ -6,6 +6,7 @@ import glob
 import sys
 from os import makedirs, remove
 from os.path import exists, join as pjoin, isfile, isdir
+import inspect
 
 import nibabel
 import pickle
@@ -18,7 +19,7 @@ from dipy.core.gradients import gradient_table
 import dipy.reconst.dti as dti
 from amico.util import PRINT, LOG, NOTE, WARNING, ERROR, get_verbose
 from pkg_resources import get_distribution
-from joblib import Parallel, delayed, cpu_count
+from joblib import Parallel, delayed, cpu_count, wrap_non_picklable_objects
 from tqdm import tqdm
 
 
@@ -35,7 +36,7 @@ def setup( lmax=12, ndirs=None ) :
     """
     if ndirs is not None:
         WARNING('"ndirs" parameter is deprecated')
-    LOG( f'\n-> Precomputing rotation matrices:' )
+    LOG( '\n-> Precomputing rotation matrices:' )
     for n in tqdm(valid_dirs(), ncols=70, bar_format='   |{bar}| {percentage:4.1f}%', disable=(get_verbose()<3)):
         amico.lut.precompute_rotation_matrices( lmax, n )
     LOG('   [ DONE ]')
@@ -47,7 +48,7 @@ class Evaluation :
     evaluation with the AMICO framework.
     """
 
-    def __init__( self, study_path, subject, output_path=None, verbose=2 ) :
+    def __init__( self, study_path, subject, output_path=None ) :
         """Setup the data structure with default values.
 
         Parameters
@@ -59,8 +60,6 @@ class Evaluation :
         output_path : string
             Optionally sets a custom full path for the output. Leave as None
             for default behaviour - output in study_path/subject/AMICO/<MODEL>
-        verbose : int
-            Possible values: 2->show all, 1->hide progress bars, 0->hide all
         """
         self.niiDWI      = None # set by "load_data" method
         self.niiDWI_img  = None
@@ -91,8 +90,6 @@ class Evaluation :
         self.set_config('DWI-SNR', None)                # SNR of DWI image: SNR = b0/sigma
         self.set_config('doDirectionalAverage', False)  # To perform the directional average on the signal of each shell
         self.set_config('parallel_jobs', -1)            # Number of jobs to be used in multithread-enabled parts of code
-        self.set_config('parallel_backend', 'loky')     # Backend to use for the joblib library
-        self.set_config('verbose', verbose)
 
     def set_config( self, key, value ) :
         self.CONFIG[ key ] = value
@@ -101,7 +98,7 @@ class Evaluation :
         return self.CONFIG.get( key )
 
 
-    def load_data( self, dwi_filename='DWI.nii', scheme_filename='DWI.scheme', mask_filename=None, b0_thr=0 ) :
+    def load_data( self, dwi_filename='DWI.nii', scheme_filename='DWI.scheme', mask_filename=None, b0_thr=0, b0_min_signal=0, replace_bad_voxels=None ) :
         """Load the diffusion signal and its corresponding acquisition scheme.
 
         Parameters
@@ -114,6 +111,10 @@ class Evaluation :
             The file name of the (optional) binary mask (default : None)
         b0_thr : float
             The threshold below which a b-value is considered a b0 (default : 0)
+        b0_min_signal : float, optional
+            Crop to zero the signal in voxels where the b0 <= b0_min_signal * mean(b0[b0>0]). (default : 0)
+        replace_bad_voxels : float, optional
+            Value to be used to fill NaN and Inf values in the signal. (default : do nothing)
         """
 
         # Loading data, acquisition scheme and mask (optional)
@@ -124,13 +125,18 @@ class Evaluation :
         if not isfile( pjoin(self.get_config('DATA_path'), dwi_filename) ):
             ERROR( 'DWI file not found' )
         self.set_config('dwi_filename', dwi_filename)
+        self.set_config('b0_min_signal', b0_min_signal)
+        self.set_config('replace_bad_voxels', replace_bad_voxels)
         self.niiDWI  = nibabel.load( pjoin(self.get_config('DATA_path'), dwi_filename) )
         self.niiDWI_img = self.niiDWI.get_data().astype(np.float32)
         hdr = self.niiDWI.header if nibabel.__version__ >= '2.0.0' else self.niiDWI.get_header()
+        if self.niiDWI_img.ndim != 4 :
+            ERROR( 'DWI file is not a 4D image' )
         self.set_config('dim', self.niiDWI_img.shape[:3])
         self.set_config('pixdim', tuple( hdr.get_zooms()[:3] ))
         PRINT('\t\t- dim    = %d x %d x %d x %d' % self.niiDWI_img.shape)
         PRINT('\t\t- pixdim = %.3f x %.3f x %.3f' % self.get_config('pixdim'))
+
         # Scale signal intensities (if necessary)
         if ( np.isfinite(hdr['scl_slope']) and np.isfinite(hdr['scl_inter']) and hdr['scl_slope'] != 0 and
             ( hdr['scl_slope'] != 1 or hdr['scl_inter'] != 0 ) ):
@@ -138,6 +144,14 @@ class Evaluation :
             self.niiDWI_img = self.niiDWI_img * hdr['scl_slope'] + hdr['scl_inter']
             PRINT('[OK]')
 
+        # Check for Nan or Inf values in raw data
+        if np.isnan(self.niiDWI_img).any() or np.isinf(self.niiDWI_img).any():
+            if replace_bad_voxels is not None:
+                WARNING(f'Nan or Inf values in the raw signal. They will be replaced with: {replace_bad_voxels}')
+                np.nan_to_num(self.niiDWI_img, copy=False, nan=replace_bad_voxels, posinf=replace_bad_voxels, neginf=replace_bad_voxels)
+            else:
+                ERROR('Nan or Inf values in the raw signal. Try using the "replace_bad_voxels" or "b0_min_signal" parameters when calling "load_data()"')
+
         PRINT('\t* Acquisition scheme')
         if not isfile( pjoin(self.get_config('DATA_path'), scheme_filename) ):
             ERROR( 'SCHEME file not found' )
@@ -163,13 +177,14 @@ class Evaluation :
             PRINT('\t\t- dim    = %d x %d x %d' % self.niiMASK_img.shape[:3])
             PRINT('\t\t- pixdim = %.3f x %.3f x %.3f' % niiMASK_hdr.get_zooms()[:3])
             if self.niiMASK.ndim != 3 :
-                ERROR( 'The provided MASK if 4D, but a 3D dataset is expected' )
+                ERROR( 'MASK file is not a 3D image' )
             if self.get_config('dim') != self.niiMASK_img.shape[:3] :
                 ERROR( 'MASK geometry does not match with DWI data' )
         else :
             self.niiMASK = None
             self.niiMASK_img = np.ones( self.get_config('dim') )
             PRINT('\t\t- not specified')
+        self.set_config('mask_filename', mask_filename)
         PRINT(f'\t\t- voxels = {np.count_nonzero(self.niiMASK_img)}')
 
         LOG( f'   [ {time.time() - tic:.1f} seconds ]' )
@@ -194,7 +209,7 @@ class Evaluation :
             else:
                 ERROR( 'No b0 volume to normalize signal with' )
             norm_factor = self.mean_b0s.copy()
-            idx = self.mean_b0s <= 0
+            idx = norm_factor <= b0_min_signal * norm_factor[norm_factor > 0].mean()
             norm_factor[ idx ] = 1
             norm_factor = 1 / norm_factor
             norm_factor[ idx ] = 0
@@ -247,6 +262,14 @@ class Evaluation :
             if self.scheme.nS != self.niiDWI_img.shape[3] :
                 ERROR( 'Scheme does not match with DWI data' )
 
+        # Check for Nan or Inf values in pre-processed data
+        if np.isnan(self.niiDWI_img).any() or np.isinf(self.niiDWI_img).any():
+            if replace_bad_voxels is not None:
+                WARNING(f'Nan or Inf values in the signal after the pre-processing. They will be replaced with: {replace_bad_voxels}')
+                np.nan_to_num(self.niiDWI_img, copy=False, nan=replace_bad_voxels, posinf=replace_bad_voxels, neginf=replace_bad_voxels)
+            else:
+                ERROR('Nan or Inf values in the signal after the pre-processing. Try using the "replace_bad_voxels" or "b0_min_signal" parameters when calling "load_data()"')
+
         LOG( f'   [ {time.time() - tic:.1f} seconds ]' )
 
 
@@ -272,11 +295,28 @@ class Evaluation :
 
     def set_solver( self, **params ) :
         """Setup the specific parameters of the solver to fit the model.
-        Dispatch to the proper function, depending on the model; a model shoudl provide a "set_solver" function to set these parameters.
+        Dispatch to the proper function, depending on the model; a model should provide a "set_solver" function to set these parameters.
+        Currently supported parameters are:
+        StickZeppelinBall:      'set_solver()' not implemented
+        CylinderZeppelinBall:   lambda1 = 0.0, lambda2 = 4.0
+        NODDI:                  lambda1 = 5e-1, lambda2 = 1e-3
+        FreeWater:              lambda1 = 0.0, lambda2 = 1e-3
+        VolumeFractions:        'set_solver()' not implemented
+        SANDI:                  lambda1 = 0.0, lambda2 = 5e-3
+        NOTE: non-existing parameters will be ignored
         """
         if self.model is None :
             ERROR( 'Model not set; call "set_model()" method first' )
-        self.set_config('solver_params', self.model.set_solver( **params ))
+
+        solver_params = list(inspect.signature(self.model.set_solver).parameters)
+        params_new = {}
+        for key in params.keys():
+            if key not in solver_params:
+                WARNING(f"Cannot find the '{key}' solver-parameter for the {self.model.name} model. It will be ignored")
+            else:
+                params_new[key] = params[key]
+
+        self.set_config('solver_params', self.model.set_solver( **params_new ))
 
 
     def generate_kernels( self, regenerate = False, lmax = 12, ndirs = 32761 ) :
@@ -356,6 +396,9 @@ class Evaluation :
         """Fit the model to the data iterating over all voxels (in the mask) one after the other.
         Call the appropriate fit() method of the actual model used.
         """
+
+        @delayed
+        @wrap_non_picklable_objects
         def fit_voxel(self, ix, iy, iz, dirs, DTI) :
             """Perform the fit in a single voxel.
             """
@@ -417,13 +460,10 @@ class Evaluation :
             n_jobs = cpu_count()
         elif n_jobs == 0 or n_jobs < -1:
             ERROR( 'Number of parallel jobs must be positive or -1' )
-        parallel_backend = self.get_config( 'parallel_backend' )
-        if parallel_backend not in ['loky','multiprocessing','threading']:
-            ERROR( f'Backend "{parallel_backend}" is not recognized by joblib' )
 
         self.set_config('fit_time', None)
         totVoxels = np.count_nonzero(self.niiMASK_img)
-        LOG( f'\n-> Fitting "{self.model.name}" model to {totVoxels} voxels:' )
+        LOG( f'\n-> Fitting "{self.model.name}" model to {totVoxels} voxels (using {n_jobs} job{"s" if n_jobs>1 else ""}):' )
 
         # setup fitting directions
         peaks_filename = self.get_config('peaks_filename')
@@ -458,8 +498,8 @@ class Evaluation :
         idx = np.arange(0, totVoxels+1, n_per_thread, dtype=np.int32)
         idx[-1] = totVoxels
 
-        estimates = Parallel(n_jobs=n_jobs, backend=parallel_backend)(
-            delayed(fit_voxel)(self, ix[i], iy[i], iz[i], DIRs[ix[i],iy[i],iz[i],:], DTI)
+        estimates = Parallel(n_jobs=n_jobs)(
+            fit_voxel(self, ix[i], iy[i], iz[i], DIRs[ix[i],iy[i],iz[i],:], DTI)
             for i in tqdm(range(totVoxels), ncols=70, bar_format='   |{bar}| {percentage:4.1f}%', disable=(get_verbose()<3))
         )
 
@@ -484,7 +524,7 @@ class Evaluation :
         if self.get_config('doSaveCorrectedDWI') :
             DWI_corrected = np.zeros(self.niiDWI.shape, dtype=np.float32)
             for i in range(totVoxels):
-                DWI_corrected[ix,iy,iz,:] = estimates[i][3]
+                DWI_corrected[ix[i],iy[i],iz[i],:] = estimates[i][3]
             self.RESULTS['DWI_corrected'] = DWI_corrected
 
 
