@@ -1,18 +1,13 @@
 from __future__ import absolute_import, division, print_function
 
 import numpy as np
-import numpy.matlib as matlib
 import scipy
-from os.path import exists, join as pjoin
-from os import remove
-import subprocess
-import tempfile
+from os.path import join as pjoin
 import amico.lut
 from tqdm import tqdm
-from dipy.core.gradients import gradient_table
-from dipy.sims.voxel import single_tensor
 import abc
-from amico.util import PRINT, LOG, NOTE, WARNING, ERROR, get_verbose
+from amico.util import PRINT, ERROR, get_verbose
+from amico.synthesis import Stick, Zeppelin, Ball, CylinderGPD, SphereGPD, Astrosticks, NODDIIntraCellular, NODDIExtraCellular, NODDIIsotropic
 
 import warnings
 warnings.filterwarnings("ignore") # needed for a problem with spams
@@ -230,28 +225,31 @@ class StickZeppelinBall( BaseModel ) :
 
 
     def generate( self, out_path, aux, idx_in, idx_out, ndirs ) :
-        scheme_high = amico.lut.create_high_resolution_scheme( self.scheme, b_scale=1 )
-        gtab = gradient_table( scheme_high.b, scheme_high.raw[:,0:3] )
+        scheme_high = amico.lut.create_high_resolution_scheme( self.scheme )
+
+        stick = Stick(scheme_high)
+        zeppelin = Zeppelin(scheme_high)
+        ball = Ball(scheme_high)
 
         nATOMS = 1 + len(self.d_perps_zep) + len(self.d_isos)
         idx = 0
         with tqdm(total=nATOMS, ncols=70, bar_format='   |{bar}| {percentage:4.1f}%', disable=(get_verbose()<3)) as progress:
             # Stick
-            signal = single_tensor( gtab, evals=[self.d_perp, self.d_perp, self.d_par] )
+            signal = stick.get_signal(self.d_par)
             lm = amico.lut.rotate_kernel( signal, aux, idx_in, idx_out, False, ndirs )
             np.save( pjoin( out_path, f'A_{idx+1:03d}.npy' ), lm )
             idx += 1
             progress.update()
             # Zeppelin(s)
             for d in self.d_perps_zep :
-                signal = single_tensor( gtab, evals=[d, d, self.d_par_zep] )
+                signal = zeppelin.get_signal(self.d_par_zep, d)
                 lm = amico.lut.rotate_kernel( signal, aux, idx_in, idx_out, False, ndirs )
                 np.save( pjoin( out_path, f'A_{idx+1:03d}.npy' ), lm )
                 idx += 1
                 progress.update()
             # Ball(s)
             for d in self.d_isos :
-                signal = single_tensor( gtab, evals=[d, d, d] )
+                signal = ball.get_signal(d)
                 lm = amico.lut.rotate_kernel( signal, aux, idx_in, idx_out, True, ndirs )
                 np.save( pjoin( out_path, f'A_{idx+1:03d}.npy' ), lm )
                 idx += 1
@@ -318,10 +316,6 @@ class CylinderZeppelinBall( BaseModel ) :
     NB: this model works only with schemes containing the full specification of
         the diffusion gradients (eg gradient strength, small delta etc).
 
-    NB: this model requires Camino to be installed and properly configured
-        in the system; in particular, the script "datasynth" must be placed
-        in your system path.
-
     References
     ----------
     .. [1] Panagiotaki et al. (2012) Compartment models of the diffusion MR signal
@@ -373,26 +367,18 @@ class CylinderZeppelinBall( BaseModel ) :
         if self.scheme.version != 1 :
             ERROR( 'This model requires a "VERSION: STEJSKALTANNER" scheme' )
 
-        scheme_high = amico.lut.create_high_resolution_scheme( self.scheme, b_scale=1E6 )
-        filename_scheme = pjoin( out_path, 'scheme.txt' )
-        np.savetxt( filename_scheme, scheme_high.raw, fmt='%15.8e', delimiter=' ', header='VERSION: STEJSKALTANNER', comments='' )
+        scheme_high = amico.lut.create_high_resolution_scheme( self.scheme )
 
-        # temporary file where to store "datasynth" output
-        filename_signal = pjoin( tempfile._get_default_tempdir(), next(tempfile._get_candidate_names())+'.Bfloat' )
+        cylinder = CylinderGPD(scheme_high)
+        zeppelin = Zeppelin(scheme_high)
+        ball = Ball(scheme_high)
 
         nATOMS = len(self.Rs) + len(self.d_perps) + len(self.d_isos)
         idx = 0
         with tqdm(total=nATOMS, ncols=70, bar_format='   |{bar}| {percentage:4.1f}%', disable=(get_verbose()<3)) as progress:
             # Cylinder(s)
             for R in self.Rs :
-                CMD = 'datasynth -synthmodel compartment 1 CYLINDERGPD %E 0 0 %E -schemefile %s -voxels 1 -outputfile %s 2> /dev/null' % ( self.d_par*1E-6, R, filename_scheme, filename_signal )
-                subprocess.call( CMD, shell=True )
-                if not exists( filename_signal ) :
-                    ERROR( 'Problems generating the signal with "datasynth"' )
-                signal  = np.fromfile( filename_signal, dtype='>f4' )
-                if exists( filename_signal ) :
-                    remove( filename_signal )
-
+                signal = cylinder.get_signal(self.d_par, R)
                 lm = amico.lut.rotate_kernel( signal, aux, idx_in, idx_out, False, ndirs )
                 np.save( pjoin( out_path, f'A_{idx+1:03d}.npy' ), lm )
                 idx += 1
@@ -400,14 +386,7 @@ class CylinderZeppelinBall( BaseModel ) :
 
             # Zeppelin(s)
             for d in self.d_perps :
-                CMD = 'datasynth -synthmodel compartment 1 ZEPPELIN %E 0 0 %E -schemefile %s -voxels 1 -outputfile %s 2> /dev/null' % ( self.d_par*1E-6, d*1e-6, filename_scheme, filename_signal )
-                subprocess.call( CMD, shell=True )
-                if not exists( filename_signal ) :
-                    ERROR( 'Problems generating the signal with "datasynth"' )
-                signal  = np.fromfile( filename_signal, dtype='>f4' )
-                if exists( filename_signal ) :
-                    remove( filename_signal )
-
+                signal = zeppelin.get_signal(self.d_par, d)
                 lm = amico.lut.rotate_kernel( signal, aux, idx_in, idx_out, False, ndirs )
                 np.save( pjoin( out_path, f'A_{idx+1:03d}.npy' ), lm )
                 idx += 1
@@ -415,14 +394,7 @@ class CylinderZeppelinBall( BaseModel ) :
 
             # Ball(s)
             for d in self.d_isos :
-                CMD = 'datasynth -synthmodel compartment 1 BALL %E -schemefile %s -voxels 1 -outputfile %s 2> /dev/null' % ( d*1e-6, filename_scheme, filename_signal )
-                subprocess.call( CMD, shell=True )
-                if not exists( filename_signal ) :
-                    ERROR( 'Problems generating the signal with "datasynth"' )
-                signal  = np.fromfile( filename_signal, dtype='>f4' )
-                if exists( filename_signal ) :
-                    remove( filename_signal )
-
+                signal = ball.get_signal(d)
                 lm = amico.lut.rotate_kernel( signal, aux, idx_in, idx_out, True, ndirs )
                 np.save( pjoin( out_path, f'A_{idx+1:03d}.npy' ), lm )
                 idx += 1
@@ -573,8 +545,11 @@ class NODDI( BaseModel ) :
 
 
     def generate( self, out_path, aux, idx_in, idx_out, ndirs ):
-        scheme_high = amico.lut.create_high_resolution_scheme( self.scheme, b_scale = 1 )
-        protocolHR = self.scheme2noddi( scheme_high )
+        scheme_high = amico.lut.create_high_resolution_scheme( self.scheme )
+
+        noddi_ic = NODDIIntraCellular(scheme_high)
+        noddi_ec = NODDIExtraCellular(scheme_high)
+        noddi_iso = NODDIIsotropic(scheme_high)
 
         nATOMS = len(self.IC_ODs)*len(self.IC_VFs) + 1
         idx = 0
@@ -582,19 +557,16 @@ class NODDI( BaseModel ) :
             # Coupled contributions
             IC_KAPPAs = 1 / np.tan(self.IC_ODs*np.pi/2)
             for kappa in IC_KAPPAs:
-                signal_ic = self.synth_meas_watson_SH_cyl_neuman_PGSE( np.array([self.dPar*1E-6, 0, kappa]), protocolHR['grad_dirs'], np.squeeze(protocolHR['gradient_strength']), np.squeeze(protocolHR['delta']), np.squeeze(protocolHR['smalldel']), np.array([0,0,1]), 0 )
-
+                signal_ic = noddi_ic.get_signal(self.dPar, kappa)
                 for v_ic in self.IC_VFs:
-                    dPerp = self.dPar*1E-6 * (1 - v_ic)
-                    signal_ec = self.synth_meas_watson_hindered_diffusion_PGSE( np.array([self.dPar*1E-6, dPerp, kappa]), protocolHR['grad_dirs'], np.squeeze(protocolHR['gradient_strength']), np.squeeze(protocolHR['delta']), np.squeeze(protocolHR['smalldel']), np.array([0,0,1]) )
-
+                    signal_ec = noddi_ec.get_signal(self.dPar, kappa, v_ic)
                     signal = v_ic*signal_ic + (1-v_ic)*signal_ec
                     lm = amico.lut.rotate_kernel( signal, aux, idx_in, idx_out, False, ndirs )
                     np.save( pjoin( out_path, f'A_{idx+1:03d}.npy') , lm )
                     idx += 1
                     progress.update()
             # Isotropic
-            signal = self.synth_meas_iso_GPD( self.dIso*1E-6, protocolHR)
+            signal = noddi_iso.get_signal(self.dIso)
             lm = amico.lut.rotate_kernel( signal, aux, idx_in, idx_out, True, ndirs )
             np.save( pjoin( out_path, f'A_{nATOMS:03d}.npy') , lm )
             progress.update()
@@ -698,395 +670,6 @@ class NODDI( BaseModel ) :
             return [v, od, fISO], dirs, x, A
 
 
-    def scheme2noddi( self, scheme ):
-        protocol = {}
-        protocol['pulseseq'] = 'PGSE'
-        protocol['schemetype'] = 'multishellfixedG'
-        protocol['teststrategy'] = 'fixed'
-        bval = scheme.b.copy()
-
-        # set total number of measurements
-        protocol['totalmeas'] = len(bval)
-
-        # set the b=0 indices
-        protocol['b0_Indices'] = np.nonzero(bval==0)[0]
-        protocol['numZeros'] = len(protocol['b0_Indices'])
-
-        # find the unique non-zero b-values
-        B = np.unique(bval[bval>0])
-
-        # set the number of shells
-        protocol['M'] = len(B)
-        protocol['N'] = np.zeros((len(B)))
-        for i in range(len(B)):
-            protocol['N'][i] = np.sum(bval==B[i])
-
-        # maximum b-value in the s/mm^2 unit
-        maxB = np.max(B)
-
-        # set maximum G = 40 mT/m
-        Gmax = 0.04
-
-        # set smalldel and delta and G
-        GAMMA = 2.675987E8
-        tmp = np.power(3*maxB*1E6/(2*GAMMA*GAMMA*Gmax*Gmax),1.0/3.0)
-        protocol['udelta'] = np.zeros((len(B)))
-        protocol['usmalldel'] = np.zeros((len(B)))
-        protocol['uG'] = np.zeros((len(B)))
-        for i in range(len(B)):
-            protocol['udelta'][i] = tmp
-            protocol['usmalldel'][i] = tmp
-            protocol['uG'][i] = np.sqrt(B[i]/maxB)*Gmax
-
-        protocol['delta'] = np.zeros(bval.shape)
-        protocol['smalldel'] = np.zeros(bval.shape)
-        protocol['gradient_strength'] = np.zeros(bval.shape)
-
-        for i in range(len(B)):
-            tmp = np.nonzero(bval==B[i])
-            for j in range(len(tmp[0])):
-                    protocol['delta'][tmp[0][j]] = protocol['udelta'][i]
-                    protocol['smalldel'][tmp[0][j]] = protocol['usmalldel'][i]
-                    protocol['gradient_strength'][tmp[0][j]] = protocol['uG'][i]
-
-        # load bvec
-        protocol['grad_dirs'] = scheme.raw[:,0:3].copy()
-
-        # make the gradient directions for b=0's [1 0 0]
-        for i in range(protocol['numZeros']):
-            protocol['grad_dirs'][protocol['b0_Indices'][i],:] = [1, 0, 0]
-
-        # make sure the gradient directions are unit vectors
-        for i in range(protocol['totalmeas']):
-            protocol['grad_dirs'][i,:] = protocol['grad_dirs'][i,:]/np.linalg.norm(protocol['grad_dirs'][i,:])
-
-        return protocol
-
-    def synth_meas_watson_SH_cyl_neuman_PGSE( self, x, grad_dirs, G, delta, smalldel, fibredir, roots ):
-        d=x[0]
-        R=x[1]
-        kappa=x[2]
-
-        l_q = grad_dirs.shape[0]
-
-        # Parallel component
-        LePar = self.cyl_neuman_le_par_PGSE(d, G, delta, smalldel)
-
-        # Perpendicular component
-        LePerp = self.cyl_neuman_le_perp_PGSE(d, R, G, delta, smalldel, roots)
-
-        ePerp = np.exp(LePerp)
-
-        # Compute the Legendre weighted signal
-        Lpmp = LePerp - LePar
-        lgi = self.legendre_gaussian_integral(Lpmp, 6)
-
-        # Compute the spherical harmonic coefficients of the Watson's distribution
-        coeff = self.watson_SH_coeff(kappa)
-        coeffMatrix = matlib.repmat(coeff, l_q, 1)
-
-        # Compute the dot product between the symmetry axis of the Watson's distribution
-        # and the gradient direction
-        #
-        # For numerical reasons, cosTheta might not always be between -1 and 1
-        # Due to round off errors, individual gradient vectors in grad_dirs and the
-        # fibredir are never exactly normal.  When a gradient vector and fibredir are
-        # essentially parallel, their dot product can fall outside of -1 and 1.
-        #
-        # BUT we need make sure it does, otherwise the legendre function call below
-        # will FAIL and abort the calculation!!!
-        #
-        cosTheta = np.dot(grad_dirs,fibredir)
-        badCosTheta = abs(cosTheta)>1
-        cosTheta[badCosTheta] = cosTheta[badCosTheta]/abs(cosTheta[badCosTheta])
-
-        # Compute the SH values at cosTheta
-        sh = np.zeros(coeff.shape)
-        shMatrix = matlib.repmat(sh, l_q, 1)
-        for i in range(7):
-            shMatrix[:,i] = np.sqrt((i+1 - .75)/np.pi)
-            # legendre function returns coefficients of all m from 0 to l
-            # we only need the coefficient corresponding to m = 0
-            # WARNING: make sure to input ROW vector as variables!!!
-            # cosTheta is expected to be a COLUMN vector.
-            tmp = np.zeros((l_q))
-            for pol_i in range(l_q):
-                tmp[pol_i] = scipy.special.lpmv(0, 2*i, cosTheta[pol_i])
-            shMatrix[:,i] = shMatrix[:,i]*tmp
-
-        E = np.sum(lgi*coeffMatrix*shMatrix, 1)
-        # with the SH approximation, there will be no guarantee that E will be positive
-        # but we need to make sure it does!!! replace the negative values with 10% of
-        # the smallest positive values
-        E[E<=0] = np.min(E[E>0])*0.1
-        E = 0.5*E*ePerp
-
-        return E
-
-    def synth_meas_watson_hindered_diffusion_PGSE( self, x, grad_dirs, G, delta, smalldel, fibredir ):
-
-        dPar = x[0]
-        dPerp = x[1]
-        kappa = x[2]
-
-        # get the equivalent diffusivities
-        dw = self.watson_hindered_diffusion_coeff(dPar, dPerp, kappa)
-
-        xh = np.array([dw[0], dw[1]])
-
-        E = self.synth_meas_hindered_diffusion_PGSE(xh, grad_dirs, G, delta, smalldel, fibredir)
-
-        return E
-
-    def cyl_neuman_le_par_PGSE( self, d, G, delta, smalldel ):
-        # Line bellow used in matlab version removed as cyl_neuman_le_par_PGSE is called from synth_meas_watson_SH_cyl_neuman_PGSE which already casts x to d, R and kappa -> x replaced by d in arguments
-        #d=x[0]
-
-        # Radial wavenumbers
-        GAMMA = 2.675987E8
-        modQ = GAMMA*smalldel*G
-        modQ_Sq = modQ*modQ
-
-        # diffusion time for PGSE, in a matrix for the computation below.
-        difftime = (delta-smalldel/3)
-
-        # Parallel component
-        LE =-modQ_Sq*difftime*d
-
-        # Compute the Jacobian matrix
-        #if(nargout>1)
-        #    % dLE/d
-        #    J = -modQ_Sq*difftime
-        #end
-
-        return LE
-
-    def cyl_neuman_le_perp_PGSE( self, d, R, G, delta, smalldel, roots ):
-
-        # When R=0, no need to do any calculation
-        if (R == 0.00):
-            LE = np.zeros(G.shape) # np.size(R) = 1
-            return LE
-        else:
-            ERROR( '"cyl_neuman_le_perp_PGSE" not yet validated for non-zero values' )
-
-    def legendre_gaussian_integral( self, Lpmp, n ):
-        if n > 6:
-            ERROR( 'The maximum value for n is 6, which corresponds to the 12th order Legendre polynomial' )
-        exact = Lpmp>0.05
-        approx = Lpmp<=0.05
-
-        mn = n + 1
-
-        I = np.zeros((len(Lpmp),mn))
-        sqrtx = np.sqrt(Lpmp[exact])
-        I[exact,0] = np.sqrt(np.pi)*scipy.special.erf(sqrtx)/sqrtx
-        dx = 1.0/Lpmp[exact]
-        emx = -np.exp(-Lpmp[exact])
-        for i in range(1,mn):
-            I[exact,i] = emx + (i-0.5)*I[exact,i-1]
-            I[exact,i] = I[exact,i]*dx
-
-        # Computing the legendre gaussian integrals for large enough Lpmp
-        L = np.zeros((len(Lpmp),n+1))
-        for i in range(n+1):
-            if i == 0:
-                L[exact,0] = I[exact,0]
-            elif i == 1:
-                L[exact,1] = -0.5*I[exact,0] + 1.5*I[exact,1]
-            elif i == 2:
-                L[exact,2] = 0.375*I[exact,0] - 3.75*I[exact,1] + 4.375*I[exact,2]
-            elif i == 3:
-                L[exact,3] = -0.3125*I[exact,0] + 6.5625*I[exact,1] - 19.6875*I[exact,2] + 14.4375*I[exact,3]
-            elif i == 4:
-                L[exact,4] = 0.2734375*I[exact,0] - 9.84375*I[exact,1] + 54.140625*I[exact,2] - 93.84375*I[exact,3] + 50.2734375*I[exact,4]
-            elif i == 5:
-                L[exact,5] = -(63./256.)*I[exact,0] + (3465./256.)*I[exact,1] - (30030./256.)*I[exact,2] + (90090./256.)*I[exact,3] - (109395./256.)*I[exact,4] + (46189./256.)*I[exact,5]
-            elif i == 6:
-                L[exact,6] = (231./1024.)*I[exact,0] - (18018./1024.)*I[exact,1] + (225225./1024.)*I[exact,2] - (1021020./1024.)*I[exact,3] + (2078505./1024.)*I[exact,4] - (1939938./1024.)*I[exact,5] + (676039./1024.)*I[exact,6]
-
-        # Computing the legendre gaussian integrals for small Lpmp
-        x2=np.power(Lpmp[approx],2)
-        x3=x2*Lpmp[approx]
-        x4=x3*Lpmp[approx]
-        x5=x4*Lpmp[approx]
-        x6=x5*Lpmp[approx]
-        for i in range(n+1):
-            if i == 0:
-                L[approx,0] = 2 - 2*Lpmp[approx]/3 + x2/5 - x3/21 + x4/108
-            elif i == 1:
-                L[approx,1] = -4*Lpmp[approx]/15 + 4*x2/35 - 2*x3/63 + 2*x4/297
-            elif i == 2:
-                L[approx,2] = 8*x2/315 - 8*x3/693 + 4*x4/1287
-            elif i == 3:
-                L[approx,3] = -16*x3/9009 + 16*x4/19305
-            elif i == 4:
-                L[approx,4] = 32*x4/328185
-            elif i == 5:
-                L[approx,5] = -64*x5/14549535
-            elif i == 6:
-                L[approx,6] = 128*x6/760543875
-
-        return L
-
-    def watson_SH_coeff( self, kappa ):
-
-        if isinstance(kappa,np.ndarray):
-            ERROR( '"watson_SH_coeff()" not implemented for multiple kappa input yet' )
-
-        # In the scope of AMICO only a single value is used for kappa
-        n = 6
-
-        C = np.zeros((n+1))
-        # 0th order is a constant
-        C[0] = 2*np.sqrt(np.pi)
-
-        # Precompute the special function values
-        sk = np.sqrt(kappa)
-        sk2 = sk*kappa
-        sk3 = sk2*kappa
-        sk4 = sk3*kappa
-        sk5 = sk4*kappa
-        sk6 = sk5*kappa
-        sk7 = sk6*kappa
-        k2 = np.power(kappa,2)
-        k3 = k2*kappa
-        k4 = k3*kappa
-        k5 = k4*kappa
-        k6 = k5*kappa
-        k7 = k6*kappa
-
-        erfik = scipy.special.erfi(sk)
-        ierfik = 1/erfik
-        ek = np.exp(kappa)
-        dawsonk = 0.5*np.sqrt(np.pi)*erfik/ek
-
-        if kappa > 0.1:
-
-            # for large enough kappa
-            C[1] = 3*sk - (3 + 2*kappa)*dawsonk
-            C[1] = np.sqrt(5)*C[1]*ek
-            C[1] = C[1]*ierfik/kappa
-
-            C[2] = (105 + 60*kappa + 12*k2)*dawsonk
-            C[2] = C[2] -105*sk + 10*sk2
-            C[2] = .375*C[2]*ek/k2
-            C[2] = C[2]*ierfik
-
-            C[3] = -3465 - 1890*kappa - 420*k2 - 40*k3
-            C[3] = C[3]*dawsonk
-            C[3] = C[3] + 3465*sk - 420*sk2 + 84*sk3
-            C[3] = C[3]*np.sqrt(13*np.pi)/64/k3
-            C[3] = C[3]/dawsonk
-
-            C[4] = 675675 + 360360*kappa + 83160*k2 + 10080*k3 + 560*k4
-            C[4] = C[4]*dawsonk
-            C[4] = C[4] - 675675*sk + 90090*sk2 - 23100*sk3 + 744*sk4
-            C[4] = np.sqrt(17)*C[4]*ek
-            C[4] = C[4]/512/k4
-            C[4] = C[4]*ierfik
-
-            C[5] = -43648605 - 22972950*kappa - 5405400*k2 - 720720*k3 - 55440*k4 - 2016*k5
-            C[5] = C[5]*dawsonk
-            C[5] = C[5] + 43648605*sk - 6126120*sk2 + 1729728*sk3 - 82368*sk4 + 5104*sk5
-            C[5] = np.sqrt(21*np.pi)*C[5]/4096/k5
-            C[5] = C[5]/dawsonk
-
-            C[6] = 7027425405 + 3666482820*kappa + 872972100*k2 + 122522400*k3  + 10810800*k4 + 576576*k5 + 14784*k6
-            C[6] = C[6]*dawsonk
-            C[6] = C[6] - 7027425405*sk + 1018467450*sk2 - 302630328*sk3 + 17153136*sk4 - 1553552*sk5 + 25376*sk6
-            C[6] = 5*C[6]*ek
-            C[6] = C[6]/16384/k6
-            C[6] = C[6]*ierfik
-
-        # for very large kappa
-        if kappa>30:
-            lnkd = np.log(kappa) - np.log(30)
-            lnkd2 = lnkd*lnkd
-            lnkd3 = lnkd2*lnkd
-            lnkd4 = lnkd3*lnkd
-            lnkd5 = lnkd4*lnkd
-            lnkd6 = lnkd5*lnkd
-            C[1] = 7.52308 + 0.411538*lnkd - 0.214588*lnkd2 + 0.0784091*lnkd3 - 0.023981*lnkd4 + 0.00731537*lnkd5 - 0.0026467*lnkd6
-            C[2] = 8.93718 + 1.62147*lnkd - 0.733421*lnkd2 + 0.191568*lnkd3 - 0.0202906*lnkd4 - 0.00779095*lnkd5 + 0.00574847*lnkd6
-            C[3] = 8.87905 + 3.35689*lnkd - 1.15935*lnkd2 + 0.0673053*lnkd3 + 0.121857*lnkd4 - 0.066642*lnkd5 + 0.0180215*lnkd6
-            C[4] = 7.84352 + 5.03178*lnkd - 1.0193*lnkd2 - 0.426362*lnkd3 + 0.328816*lnkd4 - 0.0688176*lnkd5 - 0.0229398*lnkd6
-            C[5] = 6.30113 + 6.09914*lnkd - 0.16088*lnkd2 - 1.05578*lnkd3 + 0.338069*lnkd4 + 0.0937157*lnkd5 - 0.106935*lnkd6
-            C[6] = 4.65678 + 6.30069*lnkd + 1.13754*lnkd2 - 1.38393*lnkd3 - 0.0134758*lnkd4 + 0.331686*lnkd5 - 0.105954*lnkd6
-
-        if kappa <= 0.1:
-            # for small kappa
-            C[1] = 4/3*kappa + 8/63*k2
-            C[1] = C[1]*np.sqrt(np.pi/5)
-
-            C[2] = 8/21*k2 + 32/693*k3
-            C[2] = C[2]*(np.sqrt(np.pi)*0.2)
-
-            C[3] = 16/693*k3 + 32/10395*k4
-            C[3] = C[3]*np.sqrt(np.pi/13)
-
-            C[4] = 32/19305*k4
-            C[4] = C[4]*np.sqrt(np.pi/17)
-
-            C[5] = 64*np.sqrt(np.pi/21)*k5/692835
-
-            C[6] = 128*np.sqrt(np.pi)*k6/152108775
-
-        return C
-
-    def watson_hindered_diffusion_coeff( self, dPar, dPerp, kappa ):
-
-        dw = np.zeros(2)
-        dParMdPerp = dPar - dPerp
-
-        if kappa < 1e-5:
-            dParP2dPerp = dPar + 2.*dPerp
-            k2 = kappa*kappa
-            dw[0] = dParP2dPerp/3.0 + 4.0*dParMdPerp*kappa/45.0 + 8.0*dParMdPerp*k2/945.0
-            dw[1] = dParP2dPerp/3.0 - 2.0*dParMdPerp*kappa/45.0 - 4.0*dParMdPerp*k2/945.0
-        else:
-            sk = np.sqrt(kappa)
-            dawsonf = 0.5*np.exp(-kappa)*np.sqrt(np.pi)*scipy.special.erfi(sk)
-            factor = sk/dawsonf
-            dw[0] = (-dParMdPerp+2.0*dPerp*kappa+dParMdPerp*factor)/(2.0*kappa)
-            dw[1] = (dParMdPerp+2.0*(dPar+dPerp)*kappa-dParMdPerp*factor)/(4.0*kappa)
-
-        return dw
-
-    def synth_meas_hindered_diffusion_PGSE( self, x, grad_dirs, G, delta, smalldel, fibredir ):
-
-        dPar=x[0]
-        dPerp=x[1]
-
-        # Radial wavenumbers
-        GAMMA = 2.675987E8
-        modQ = GAMMA*smalldel*G
-        modQ_Sq = np.power(modQ,2.0)
-
-        # Angles between gradient directions and fibre direction.
-        cosTheta = np.dot(grad_dirs,fibredir)
-        cosThetaSq = np.power(cosTheta,2.0)
-        sinThetaSq = 1.0-cosThetaSq
-
-        # b-value
-        bval = (delta-smalldel/3.0)*modQ_Sq
-
-        # Find hindered signals
-        E=np.exp(-bval*((dPar - dPerp)*cosThetaSq + dPerp))
-
-        return E
-
-    def synth_meas_iso_GPD( self, d, protocol ):
-        if protocol['pulseseq'] != 'PGSE' and protocol['pulseseq'] != 'STEAM':
-            ERROR( 'synth_meas_iso_GPD() : Protocol %s not translated from NODDI matlab code yet' % protocol['pulseseq'] )
-
-        GAMMA = 2.675987E8
-        modQ = GAMMA*protocol['smalldel'].transpose()*protocol['gradient_strength'].transpose()
-        modQ_Sq = np.power(modQ,2)
-        difftime = protocol['delta'].transpose()-protocol['smalldel']/3.0
-        return np.exp(-difftime*modQ_Sq*d)
-
-
 
 class FreeWater( BaseModel ) :
     """Implements the Free-Water model.
@@ -1165,15 +748,17 @@ class FreeWater( BaseModel ) :
 
 
     def generate( self, out_path, aux, idx_in, idx_out, ndirs ) :
-        scheme_high = amico.lut.create_high_resolution_scheme( self.scheme, b_scale=1 )
-        gtab = gradient_table( scheme_high.b, scheme_high.raw[:,0:3] )
+        scheme_high = amico.lut.create_high_resolution_scheme( self.scheme )
+
+        zeppelin = Zeppelin(scheme_high)
+        ball = Ball(scheme_high)
 
         nATOMS = len(self.d_perps) + len(self.d_isos)
         idx = 0
         with tqdm(total=nATOMS, ncols=70, bar_format='   |{bar}| {percentage:4.1f}%', disable=(get_verbose()<3)) as progress:
             # Tensor compartment(s)
             for d in self.d_perps :
-                signal = single_tensor( gtab, evals=[d, d, self.d_par] )
+                signal = zeppelin.get_signal(self.d_par, d)
                 lm = amico.lut.rotate_kernel( signal, aux, idx_in, idx_out, False, ndirs )
                 np.save( pjoin( out_path, f'A_{idx+1:03d}.npy' ), lm )
                 idx += 1
@@ -1181,7 +766,7 @@ class FreeWater( BaseModel ) :
 
             # Isotropic compartment(s)
             for d in self.d_isos :
-                signal = single_tensor( gtab, evals=[d, d, d] )
+                signal = ball.get_signal(d)
                 lm = amico.lut.rotate_kernel( signal, aux, idx_in, idx_out, True, ndirs )
                 np.save( pjoin( out_path, f'A_{idx+1:03d}.npy' ), lm )
                 idx += 1
@@ -1306,6 +891,7 @@ class VolumeFractions( BaseModel ) :
         ERROR( 'Not implemented' )
 
 
+
 class SANDI( BaseModel ) :
     """Implements the SANDI model [1].
 
@@ -1317,10 +903,6 @@ class SANDI( BaseModel ) :
     NB: this model works only with direction-averaged signal and schemes containing the full specification of
         the diffusion gradients (eg gradient strength, small delta etc).
 
-    NB: this model requires Camino to be installed and properly configured
-        in the system; in particular, the script "datasynth" must be placed
-        in your system path.
-
     References
     ----------
     .. [1] Palombo, Marco, et al. "SANDI: a compartment-based model for non-invasive apparent soma and neurite imaging by diffusion MRI." Neuroimage 215 (2020): 116835.
@@ -1369,51 +951,32 @@ class SANDI( BaseModel ) :
         if self.scheme.version != 1 :
             ERROR( 'This model requires a "VERSION: STEJSKALTANNER" scheme' )
 
-        scheme_high = amico.lut.create_high_resolution_scheme( self.scheme, b_scale=1E6 )
-        filename_scheme = pjoin( out_path, 'scheme.txt' )
-        np.savetxt( filename_scheme, scheme_high.raw, fmt='%15.8e', delimiter=' ', header='VERSION: STEJSKALTANNER', comments='' )
+        scheme_high = amico.lut.create_high_resolution_scheme( self.scheme )
 
-        # temporary file where to store "datasynth" output
-        filename_signal = pjoin( tempfile._get_default_tempdir(), next(tempfile._get_candidate_names())+'.Bfloat' )
+        sphere = SphereGPD(scheme_high)
+        astrosticks = Astrosticks(scheme_high)
+        ball = Ball(scheme_high)
 
         nATOMS = len(self.Rs) + len(self.d_in) + len(self.d_isos)
         idx = 0
         with tqdm(total=nATOMS, ncols=70, bar_format='   |{bar}| {percentage:4.1f}%', disable=(get_verbose()<3)) as progress:
             # Soma = SPHERE
             for R in self.Rs :
-                CMD = 'datasynth -synthmodel compartment 1 SPHEREGPD %E %E -schemefile %s -voxels 1 -outputfile %s 2> /dev/null' % ( self.d_is*1E-6, R, filename_scheme, filename_signal )
-                subprocess.call( CMD, shell=True )
-                if not exists( filename_signal ) :
-                    ERROR( 'Problems generating the signal with "datasynth"' )
-                signal  = np.fromfile( filename_signal, dtype='>f4' )
-                if exists( filename_signal ) :
-                    remove( filename_signal )
+                signal = sphere.get_signal(self.d_is, R)
                 lm = amico.lut.rotate_kernel( signal, aux, idx_in, idx_out, True, ndirs )
                 np.save( pjoin( out_path, f'A_{idx+1:03d}.npy' ), lm )
                 idx += 1
                 progress.update()
             # Neurites = ASTRO STICKS
             for d in self.d_in :
-                CMD = 'datasynth -synthmodel compartment 1 ASTROSTICKS %E -schemefile %s -voxels 1 -outputfile %s 2> /dev/null' % ( d*1e-6, filename_scheme, filename_signal )
-                subprocess.call( CMD, shell=True )
-                if not exists( filename_signal ) :
-                    ERROR( 'Problems generating the signal with "datasynth"' )
-                signal  = np.fromfile( filename_signal, dtype='>f4' )
-                if exists( filename_signal ) :
-                    remove( filename_signal )
+                signal = astrosticks.get_signal(d)
                 lm = amico.lut.rotate_kernel( signal, aux, idx_in, idx_out, True, ndirs )
                 np.save( pjoin( out_path, f'A_{idx+1:03d}.npy' ), lm )
                 idx += 1
                 progress.update()
             # Extra-cellular = BALL
             for d in self.d_isos :
-                CMD = 'datasynth -synthmodel compartment 1 BALL %E -schemefile %s -voxels 1 -outputfile %s 2> /dev/null' % ( d*1e-6, filename_scheme, filename_signal )
-                subprocess.call( CMD, shell=True )
-                if not exists( filename_signal ) :
-                    ERROR( 'Problems generating the signal with "datasynth"' )
-                signal  = np.fromfile( filename_signal, dtype='>f4' )
-                if exists( filename_signal ) :
-                    remove( filename_signal )
+                signal = ball.get_signal(d)
                 lm = amico.lut.rotate_kernel( signal, aux, idx_in, idx_out, True, ndirs )
                 np.save( pjoin( out_path, f'A_{idx+1:03d}.npy' ), lm )
                 idx += 1
